#include "include.h"

#ifndef ENGINE_H
#define ENGINE_H


class Engine
{
public:
	void init(void *param1, void *param2);
	void destroy();
	void quit();

	void render();
	void render_scene(bool lights);
	void draw_plane(plane_t &p, bool lights);

	void draw_points();

	void step();
	void dynamics();

	bool collision_detect(RigidBody &body);
	bool body_collision(RigidBody &body);

	void resize(int width, int height);
	bool mousepos(int x, int y, int deltax, int deltay);
	void keypress(char *key, bool pressed);


private:
	vector<RigidBody *>	entity_list;
	void	*param1;
	void	*param2;
	button_t keyboard;
	Graphics gfx;
	int width;
	int height;
	float zNear;
	float zFar;


	Model model;
	matrix4 transformation;
	matrix4 projection;

	unsigned int quad_tex;
	unsigned int depth_tex;
	unsigned int fbo;
	unsigned int fbo_width;
	unsigned int fbo_height;


	GlobalShader shader;
	Frame light_frame;
	Frame camera_frame;

	RadiantMap radmap;




#ifdef WIN32
	HWND	hwnd;
	HDC	hdc;
#endif

#ifdef __linux__
	Display	*display;
	Window	window;
#endif

#ifdef __MACH__
	WindowPtr		window;
	AGLContext		agl_context;
#endif
};

#endif

